unit debug_logger;

{$mode ObjFPC}{$H+}

interface

uses
	SysUtils, Classes,
	Windows,
	comparators, handletables,
	hooks,
	wc3main, wc3jassvm, jassvm_extension, logging, stringbuilderhack;

procedure OnNativeCall(Instance: PInstance; hName: Dword); stdcall;
procedure OnCall(Instance: PInstance; hName: Dword); stdcall;
procedure OnReturn(Instance: PInstance); stdcall;

implementation

type
	TJumpTable = array [0..41] of Pointer;
	PJumpTable = ^TJumpTable;
	TFunctionTable = array of Boolean;
	TFunctionTableComparator = specialize TPointerComparator<TFunctionTable>;
	TFunctionTables = specialize THandleTableM<TFunctionTable, TFunctionTableComparator>;

	PPScriptData = ^PScriptData;

const
	DEFAULT_LOG_FILE_PATH = './jass_logger.log';
	DEFAULT_BLACKLIST_PATH = './jass_logger.blacklist';

var
	_pJumpTable: PJumpTable;
	JumpTableCopy: TJumpTable;
	orig_Compile: function ({Files: PScriptFile; Callback: TScriptCompileCallbackWar3;} ErrorCallbackParam: Dword; RiskLevel: Dword): TScriptHandle; stdcall; // fastcall

	LogFilePath: String = DEFAULT_LOG_FILE_PATH;
	Logger: TLogger;

	Blacklist: TStringList;
	FunctionTables: TFunctionTables;
	IgnoreLevel: Integer = 0;

	StringBuilder: TAnsiStringBuilderHack;

function ScriptDataValueToString(Instance: PInstance; Data: PScriptData): String;
begin
	case Data^.InitializedType of
		stVoid: Result:= 'nothing';
		stRetAddr: Result:= Format('return address %X', [Data^.Value]);
		stNull: Result:= 'null';
		stCode: Result:= Format('function %s', [GetCodeName(
			Instance,
			PInstruction(Data^.Value)
		)]);
		stInteger: Result:= IntToStr(PInt32(@Data^.Value)^); //Result:= 'int';
		stReal: Result:= FloatToStr(PSingle(@Data^.Value)^); //Result:= 'real';
		stString: Result:= Format('"%s"', [Instance^.StringTable[Data^.Value].Content]);
		stHandle: Result:= '0x' + IntToHex(Data^.Value);
		stBoolean: Result:= BoolToStr(LongBool(Data^.Value), 'true', 'false');
		stArrayOfInteger,
		stArrayOfReal,
		stArrayOfString,
		stArrayOfHandle,
		stArrayOfBoolean: Result:= Format('array at 0x%.8X', [Data^.Value]);
		else Result:= Format('unknown(0x%X)', [Data^.Value]);
	end;
end;

procedure FormatArguments(Instance: PInstance; LastArgument: PPScriptData; hName: Dword; AStringBuilder: TAnsiStringBuilderHack);
var
	hScript: TScriptHandle;
	Functions: TFuncTable;
	Count: Integer;
	i: Integer;
	FirstArgument: PPScriptData;
	s: String;
begin
	hScript:= GetInstanceScriptHandle(Instance);
	Functions:= GetScriptFunctions(hScript);
	Count:= Functions[hName].ArgsCount;
	FirstArgument:= LastArgument - (Count - 1);

	AStringBuilder.Append('(');
	for i:= 0 to Count - 1 do begin
		s:= ScriptDataValueToString(Instance, FirstArgument[i]);
		AStringBuilder.Append(s);

		if i < (Count - 1)
			then AStringBuilder.Append(', ');
	end;
	AStringBuilder.Append(')');
end;

procedure OnCall(Instance: PInstance; hName: Dword); stdcall;
var
	Name: PChar;
	LoggedFunctions: TFunctionTable;
	Arguments: PPScriptData;
	hScript: TScriptHandle;
begin
	if IgnoreLevel = 0 then begin
		hScript:= GetInstanceScriptHandle(Instance);
		LoggedFunctions:= FunctionTables.GetValue(hScript);

		if LoggedFunctions[hName] then begin
			Name:= Instance^.Script^.Strings^[hName];
			StringBuilder.Length:= 0;
			StringBuilder.Append(Name);

			// The new frame has not yet been created, so look for arguments in the current one.
			// The return address has not yet been pushed, so don't skip them.
			with Instance^.Frames.First^.Stack do begin
				Arguments:= @Items[Length - 1];
			end;

			FormatArguments(Instance, Arguments, hName, StringBuilder);
			Logger.DebugLnEnter(@StringBuilder.Data[0], StringBuilder.Length);
		end
		else Inc(IgnoreLevel);
	end
	else begin
		Inc(IgnoreLevel);
	end;
end;

procedure OnNativeCall(Instance: PInstance; hName: Dword); stdcall;
var
	Name: PChar;
	Arguments: PPScriptData;
begin
	Name:= Instance^.Script^.Strings^[hName];
	StringBuilder.Length:= 0;
	StringBuilder.Append(Name);

	with Instance^.Frames.First^.Stack do begin
		Arguments:= @Items[Length - 1];
	end;

	FormatArguments(Instance, Arguments, hName, StringBuilder);
	Logger.DebugLn(@StringBuilder.Data[0], StringBuilder.Length);
end;

procedure op_call_jump_case; assembler; nostackframe;
const
	OP_CALL_JUMP_CASE: Integer = Integer(opCall) - 2;
asm
	pushad
	push eax
	push esi
	call OnCall
	popad
	mov ecx, OP_CALL_JUMP_CASE
	jmp JumpTableCopy[ecx*4]
end;

procedure OnReturn(Instance: PInstance); stdcall;
var
	Frame: PFrame;
	IsLastFrame: Boolean;
begin
	Frame:= Instance^.Frames.First;
	if Frame <> nil then begin
		IsLastFrame:= Instance^.Frames.GetNextItem(Frame) = nil;
		if not Frame^.IsFirstFrame and not IsLastFrame then begin
			if IgnoreLevel = 0
				then Logger.DebugLnExit()
				else Dec(IgnoreLevel);
		end;
	end;
end;

procedure op_return_jump_case; assembler; nostackframe;
const
	OP_RETURN_JUMP_CASE: Integer = Integer(opReturn) - 2;
asm
	push esi
	call OnReturn
	mov ecx, OP_RETURN_JUMP_CASE
	jmp JumpTableCopy[ecx*4]
end;

function OperationCaseIndex(id: TOperationCode): Integer;
begin
	Result:= Integer(id) - 2;
end;

procedure HookInstruction(id: TOperationCode; Target: Pointer);
begin
	if not WriteProcessMemory(
		GetCurrentProcess(),
		@_pJumpTable^[OperationCaseIndex(id)],
		@Target,
		SizeOf(Target),
		nil
	)
	then Logger.DebugLn('Failed to hook instruction %d.', [id]);
end;

procedure PatchTable;
begin
    MoveMemory(@JumpTableCopy, _pJumpTable, SizeOf(TJumpTable));
	HookInstruction(opCall, @op_call_jump_case);
	HookInstruction(opReturn, @op_return_jump_case);
end;

function GenerateFunctionTable(hScript: TScriptHandle): TFunctionTable;
var
	pScript: PJassScript;
	Name: String;
	Index: Dword;
begin
	pScript:= GetJassThreadLocal()^.Scripts[hScript];
	SetLength(Result, pScript^.Strings^.Length);
	FillByte(Result[0], Length(Result) * SizeOf(Boolean), Byte(True));

	for Name in Blacklist do begin
		if pScript^.Strings^.GetItemIndex(PChar(Name), Index) then begin
			Result[Index]:= False;
		end;
	end;
end;

procedure InitScriptFunctionsTable(hScript: TScriptHandle);
var
	Table: TFunctionTable;
begin
	Table:= GenerateFunctionTable(hScript);
	FunctionTables.InsertValue(hScript, Table);
end;

function hook_Compile({Files: PScriptFile; Callback: TScriptCompileCallbackWar3;} ErrorCallbackParam: Dword; RiskLevel: Dword): TScriptHandle; stdcall; // fastcall
begin
	Result:= orig_Compile(ErrorCallbackParam, RiskLevel);
	InitScriptFunctionsTable(Result);
end;

function FilterCommentedEntries(Sender: TObject; const s: String): Boolean;
begin
	Result:= s.StartsWith('#');
end;

function ToObjectProcedure(Callback: Pointer): TMethod;
begin
	Result.Code:= Callback;
	Result.Data:= nil;
end;

type
	TStringFilterFunction = function (Sender: TObject; const s: String): Boolean;

function AsStringFilter(Callback: TStringFilterFunction): TStringsFilterMethod;
begin
	Result:= TStringsFilterMethod(ToObjectProcedure(Callback))
end;

procedure LoadBlacklist;
begin
	try
		Blacklist.LoadFromFile(DEFAULT_BLACKLIST_PATH);
	except
		on e: EFOpenError do begin
			Blacklist.Clear;
		end;
	end;
	Blacklist.Filter(AsStringFilter(@FilterCommentedEntries));
end;

function ExtractParameter(Text: String; Name: String; out Value: String): Boolean;
begin
	Result:= False;

	if Text.StartsWith(Name, True) then begin
		Value:= Copy(Text, Length(Name) + 1, Length(Text));
		Result:= True;
	end;
end;

procedure ParseCommandLine;
var
	i: Integer;
	Value: String;
begin
	for i:= 1 to ParamCount do begin
		if ExtractParameter(ParamStr(i), '--jass-logger-file=', Value)
			then LogFilePath:= Value;
	end;
end;

procedure OnInit;
begin
	Pointer(_pJumpTable):= Pointer(GameDll + $45F808);
	//PatchTable;

	if not Hook(
		Pointer(_Compile),
		@hook_Compile,
		Pointer(orig_Compile)
	)
	then Logger.DebugLn('Failed to hook compiler.');
end;

begin
	wc3main.SubscribeOnInit(@OnInit);
	ParseCommandLine;
	Logger:= TLogger.Create(LogFilePath);

	Blacklist:= TStringList.Create;
	FunctionTables:= TFunctionTables.Create;
	LoadBlacklist;

	StringBuilder:= TAnsiStringBuilderHack.Create;
end.


