unit jassvm_extension;

{$mode ObjFPC}{$H+}

interface

uses
	wc3jassvm;

type
	TFuncData = record
		ArgsCount: Integer;
	end;

	TFuncTable = array of TFuncData;

function GetCodeName(Instance: PInstance; Code: PInstruction): String;
function GetScriptFunctions(hScript: TScriptHandle): TFuncTable;
function GetInstanceScriptHandle(Instance: PInstance): TScriptHandle;

implementation

uses
	SysUtils,
	hooks,
	comparators, handletables,
	wc3main;

type
	TScriptToFuncTable = specialize THandleTableM<TFuncTable, specialize TPointerComparator<TFuncTable>>;
	TInstanceToScriptTable = specialize THandleTable<TScriptHandle>;

	{ TCodeIterator }

	TCodeIterator = object
	private
		FScript: PJassScript;
		FInstructionsCount: Integer;
		FCurrentIndex: Integer;
	public
		function GetCurrent: PInstruction;
		function MoveNext: Boolean;
		procedure Reset;
		property Current: PInstruction read GetCurrent;
		constructor Init(AScript: PJassScript);
		function GetEnumerator: TCodeIterator;
	end;

var
	InstanceToScriptTable: TInstanceToScriptTable;
	ScriptToFuncTable: TScriptToFuncTable;
	orig_RegisterInstance: function ({Instance: PInstance; DesiredSlot: TInstanceHandle}): TInstanceHandle; // fastcall
	orig_Compile: function ({Files: PScriptFile; Callback: TScriptCompileCallbackWar3;} ErrorCallbackParam: Dword; RiskLevel: Dword): TScriptHandle; stdcall; // fastcall

function GetCodeName(Instance: PInstance; Code: PInstruction): String;
var
	FunctionLabel: PInstruction;
begin
	if Code <> nil then begin
		FunctionLabel:= Code - 1;
		if TOperationCode(FunctionLabel^.id) = opFunction then begin
			Result:= Instance^.Script^.Strings^[FunctionLabel^.Parameter];
		end
		else begin
			Result:= '0x' +  HexStr(Code);
		end;
	end
	else begin
		Result:= 'nullptr';
	end;
end;

function FindScriptHandle(Script: PJassScript): TScriptHandle;
var
	i: Dword;
	Environment: PJassThreadLocal;
begin
	Result:= 0;

	Environment:= GetJassThreadLocal();
	for i:= 0 to Environment^.Scripts.Length - 1 do begin
		if Script = Environment^.Scripts[i] then begin
			Result:= i;
			break;
		end;
	end;
end;

function hook_RegisterInstance({Instance: PInstance; DesiredSlot: TInstanceHandle}): TInstanceHandle; // fastcall
var
	Instance: PInstance;
begin
	asm mov Instance, ecx end;

	Result:= orig_RegisterInstance();
	InstanceToScriptTable[Result]:= FindScriptHandle(Instance^.Script);
end;

procedure InitScriptFuncTable(hScript: TScriptHandle);
var
	Environment: PJassThreadLocal;
	Table: TFuncTable;
	Script: PJassScript;
	Instruction: PInstruction;
	Instructions: TCodeIterator;
	CurrentFunction: Dword = 0;
	Native: PNativeFunc;
	StringIndex: Dword;
begin
	Environment:= GetJassThreadLocal();
	Script:= Environment^.Scripts[hScript];
	Instructions.Init(Script);

	SetLength(Table, Script^.Strings^.Length);

	for Instruction in Instructions do begin
		case TOperationCode(Instruction^.id) of
			opEndProgram: break;
			opFunction: CurrentFunction:= Instruction^.Parameter;
			opArgument: Inc(Table[CurrentFunction].ArgsCount);
		end;
	end;

	for Native in Environment^.Natives do begin
		if Script^.Strings^.GetItemIndex(
			Native^.HashKeyData.Key,
			StringIndex
		) then begin
			Table[StringIndex].ArgsCount:= Native^.ArgsCount;
		end;
	end;

	ScriptToFuncTable[hScript]:= Table;
end;

function hook_Compile({Files: PScriptFile; Callback: TScriptCompileCallbackWar3;} ErrorCallbackParam: Dword; RiskLevel: Dword): TScriptHandle; stdcall; // fastcall
begin
	Result:= orig_Compile(ErrorCallbackParam, RiskLevel);
	InitScriptFuncTable(Result);
end;

procedure OnInit;
begin
	if not Hook(
		Pointer(_RegisterInstance),
		@hook_RegisterInstance,
		Pointer(orig_RegisterInstance)
	)
	then raise Exception.Create('Failed to hook RegisterInstance.');

	if not Hook(
		Pointer(_Compile),
		@hook_Compile,
		Pointer(orig_Compile)
	)
	then raise Exception.Create('Failed to hook compiler.');
end;

function GetScriptFunctions(hScript: TScriptHandle): TFuncTable;
begin
	Result:= ScriptToFuncTable[hScript];
end;

function GetInstanceScriptHandle(Instance: PInstance): TScriptHandle;
begin
	Result:= InstanceToScriptTable[Instance^.Handle];
end;

{ TCodeIterator }

function TCodeIterator.GetCurrent: PInstruction;
begin
	Result:= @FScript^.Code[FCurrentIndex];
end;

function TCodeIterator.MoveNext: Boolean;
begin
	if FCurrentIndex < (FInstructionsCount - 1) then begin
		Inc(FCurrentIndex);
		Result:= True;
	end else begin
		Result:= False;
	end;
end;

procedure TCodeIterator.Reset;
begin
	FCurrentIndex:= -1;
end;

constructor TCodeIterator.Init(AScript: PJassScript);
begin
	FScript:= AScript;
	FInstructionsCount:= FScript^.CodeSize div SizeOf(TInstruction);
	Reset;
end;

function TCodeIterator.GetEnumerator: TCodeIterator;
begin
	Result:= Self;
end;

initialization
	InstanceToScriptTable:= TInstanceToScriptTable.Create;
	ScriptToFuncTable:= TScriptToFuncTable.Create;
	wc3main.SubscribeOnInit(@OnInit);

end.


